@@ -4,13 +4,15 @@
  * created by Marc Ewing.
  * Currently maintained by Andrew G. Morgan <morgan@parc.power.net>
  *
- * $Id: pam_handlers.c,v 1.15 1997/01/04 20:07:56 morgan Exp $
+ * $Id: pam_handlers.c,v 1.16 1997/02/15 15:57:32 morgan Exp $
  *
  */
 
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
+#include <sys/types.h>
+#include <sys/stat.h>
 #include <dlfcn.h>
 #include <fcntl.h>
 #include <unistd.h>
@@ -20,100 +22,57 @@
 #define BUF_SIZE      1024
 #define MODULE_CHUNK  4
 
-int _pam_assemble_line(FILE *f, char *buf, int buf_len);
+static int _pam_assemble_line(FILE *f, char *buf, int buf_len);
 
-void _pam_free_handlers_aux(struct handler **hp);
+static void _pam_free_handlers_aux(struct handler **hp);
 
-int _pam_add_handler(pam_handle_t *pamh
+static int _pam_add_handler(pam_handle_t *pamh
 		     , int must_fail, int other, int type
 		     , int control_flag, const char *mod_path
 		     , int argc, char **argv, int argvlen);
 
-/* Values for type */
+/* Values for module type */
+
 #define PAM_T_AUTH    1
 #define PAM_T_SESS    2
 #define PAM_T_ACCT    4
 #define PAM_T_PASS    8
 
-/* Parse config file, allocate handler structures, dlopen() */
-int _pam_init_handlers(pam_handle_t *pamh)
+static int _pam_parse_conf_file(pam_handle_t *pamh, FILE *f
+				, const char *known_service /* specific file */
+    )
 {
-    FILE *f;
     char buf[BUF_SIZE];
-    char *mod_path;
-    int x, module_type, control_flag;
-    int other;
-    char *service_name = pamh->service_name;
-    int res;
-#ifdef DEBUG
-    int y;
-#endif
-
-    int argc;
-    char **argv;
-    int argvlen;
-
-    D(("_pam_init_handlers called"));
-    IF_NO_PAMH("_pam_init_handlers",pamh,PAM_SYSTEM_ERR);
-
-    /* Return immediately if everything is already loaded */
-    if (pamh->handlers.handlers_loaded) {
-	return PAM_SUCCESS;
-    }
-
-    D(("_pam_init_handlers: initializing"));
-    
-    /* First clean the service structure */
-
-    _pam_free_handlers(pamh);
-    if (! pamh->handlers.module) {
-	if ((pamh->handlers.module =
-	     malloc(MODULE_CHUNK * sizeof(struct loaded_module))) == NULL) {
-	    _pam_log_error("_pam_init_handlers: no memory loading module");
-	    return PAM_BUF_ERR;
-	}
-	pamh->handlers.modules_allocated = MODULE_CHUNK;
-	pamh->handlers.modules_used = 0;
-    }
-    
-    /* Now parse the config file and add handlers */
-
-    if ((f = fopen(PAM_CONFIG, "r")) == NULL) {
-	_pam_log_error("_pam_init_handlers: could not open " PAM_CONFIG );
-	return PAM_ABORT;
-    }
-
-#ifdef PAM_LOCKING
-    {
-	 int fd_tmp;
-
-	 if ((fd_tmp = open( PAM_LOCK_FILE, O_RDONLY )) != -1) {
-	      _pam_log_error("_pam_init_handlers: PAM lockfile ("
-			     PAM_LOCK_FILE ") exists - aborting");
-	      (void) close(fd_tmp);
-#ifdef PAM_FAIL_DELAY_ON
-	      _pam_start_timer(pamh);
-	      pam_fail_delay(pamh, 5000000);
-	      _pam_await_timer(pamh, PAM_ABORT);
-#else
-	      sleep(5);
-#endif /* PAM_FAIL_DELAY_ON */
-	      return PAM_ABORT;
-	 }
-    }
-#endif /* PAM_LOCKING */
-
+    int x;                    /* read a line from the FILE *f ? */
+    /*
+     * read a line from the configuration (FILE *) f
+     */
     while ((x = _pam_assemble_line(f, buf, BUF_SIZE)) > 0) {
 	char *tok, *nexttok=NULL;
 	const char *this_service;
+	const char *mod_path;
+	int module_type, control_flag;
+	int other;            /* set if module is for PAM_DEFAULT_SERVICE */
+	int res;              /* module added successfully? */
 	int must_fail=0;      /* a badly formatted line must fail when used */
+	int argc;
+	char **argv;
+	int argvlen;
 
 	D(("_pam_init_handler: LINE: %s", buf));
-	this_service = tok = _pam_StrTok(buf, " \n\t", &nexttok);
-	other = !_pam_strCMP(tok, "OTHER");
-	if (!_pam_strCMP(tok, service_name) || other) {
-	    /* This is the service we are looking for */
-	    D(("_pam_init_handlers: Found pam.conf entry for: %s", tok));
+	if (known_service != NULL) {
+	    nexttok = buf;
+	    /* No service field: all lines are for the known service. */
+	    this_service = known_service;
+	} else {
+	    this_service = tok = _pam_StrTok(buf, " \n\t", &nexttok);
+	}
+
+	other = !_pam_strCMP(this_service, PAM_DEFAULT_SERVICE);
+	if (!_pam_strCMP(this_service, pamh->service_name) || other) {
+	    /* This is a service we are looking for */
+	    D(("_pam_init_handlers: Found PAM config entry for: %s"
+	       , this_service));
 
 	    tok = _pam_StrTok(NULL, " \n\t", &nexttok);
 	    if (!_pam_strCMP("auth", tok)) {
@@ -132,9 +91,9 @@ int _pam_init_handlers(pam_handle_t *pamh)
 		module_type = PAM_T_AUTH;                  /* most sensitive */
 		must_fail = 1; /* install as normal but fail when dispatched */
 	    }
-	    D(("Using %spam.conf entry: %s", must_fail?"BAD ":"", tok));
+	    D(("Using %s config entry: %s", must_fail?"BAD ":"", tok));
 
-	    /* accept "service name" or OTHER of this module_type */
+	    /* accept "service name" or PAM_DEFAULT_SERVICE modules */
 	    
 	    tok = _pam_StrTok(NULL, " \n\t", &nexttok);
 	    if (!_pam_strCMP("required", tok)) {
@@ -183,11 +142,17 @@ int _pam_init_handlers(pam_handle_t *pamh)
 	    }
 
 #ifdef DEBUG
-	    D(("CONF%s: %s %d %d %s %d", must_fail?"<*will fail*>":""
-	       , service_name, module_type
-	       , control_flag, mod_path, argc));
-	    for (y = 0; y < argc; y++) {
-		D(("CONF: %s", argv[y]));
+	    {
+		int y;
+
+		D(("CONF%s: %s%s %d %d %s %d"
+		   , must_fail?"<*will fail*>":""
+		   , this_service, other ? "(backup)":""
+		   , module_type, control_flag
+		   , mod_path, argc));
+		for (y = 0; y < argc; y++) {
+		    D(("CONF: %s", argv[y]));
+		}
 	    }
 #endif
 
@@ -202,12 +167,151 @@ int _pam_init_handlers(pam_handle_t *pamh)
 	}
     }
 
-    D(("closing configuration file"));
-    fclose(f);
+    return ( (x < 0) ? PAM_ABORT:PAM_SUCCESS );
+}
 
-    if (x < 0) {
+/* Parse config file, allocate handler structures, dlopen() */
+int _pam_init_handlers(pam_handle_t *pamh)
+{
+    FILE *f;
+    int retval;
+
+    D(("_pam_init_handlers called"));
+    IF_NO_PAMH("_pam_init_handlers",pamh,PAM_SYSTEM_ERR);
+
+    /* Return immediately if everything is already loaded */
+    if (pamh->handlers.handlers_loaded) {
+	return PAM_SUCCESS;
+    }
+
+    D(("_pam_init_handlers: initializing"));
+    
+    /* First clean the service structure */
+
+    _pam_free_handlers(pamh);
+    if (! pamh->handlers.module) {
+	if ((pamh->handlers.module =
+	     malloc(MODULE_CHUNK * sizeof(struct loaded_module))) == NULL) {
+	    _pam_log_error("_pam_init_handlers: no memory loading module");
+	    return PAM_BUF_ERR;
+	}
+	pamh->handlers.modules_allocated = MODULE_CHUNK;
+	pamh->handlers.modules_used = 0;
+    }
+
+    if (pamh->service_name == NULL) {
+	return PAM_BAD_ITEM;                /* XXX - better error? */
+    }
+
+#ifdef PAM_LOCKING
+    /* Is the PAM subsystem locked? */
+    {
+	 int fd_tmp;
+
+	 if ((fd_tmp = open( PAM_LOCK_FILE, O_RDONLY )) != -1) {
+	      _pam_log_error("_pam_init_handlers: PAM lockfile ("
+			     PAM_LOCK_FILE ") exists - aborting");
+	      (void) close(fd_tmp);
+#ifdef PAM_FAIL_DELAY_ON
+	      _pam_start_timer(pamh);
+	      pam_fail_delay(pamh, 5000000);
+	      _pam_await_timer(pamh, PAM_ABORT);
+#else
+	      sleep(5);     /* to avoid swamping the system with requests */
+#endif /* PAM_FAIL_DELAY_ON */
+	      return PAM_ABORT;
+	 }
+    }
+#endif /* PAM_LOCKING */
+
+    /*
+     * Now parse the config file(s) and add handlers
+     */
+    {
+	struct stat test_d;
+	
+	/* Is there a PAM_CONFIG_D directory? */
+	if ( stat(PAM_CONFIG_D, &test_d) == 0 && S_ISDIR(test_d.st_mode) ) {
+	    char *filename;
+	    int read_something=0;
+
+	    D(("searching " PAM_CONFIG_D " for config files"));
+	    filename = malloc(sizeof(PAM_CONFIG_DF)
+			      +strlen(pamh->service_name));
+	    if (filename == NULL) {
+		_pam_log_error("_pam_init_handlers: no memory; service %s"
+			       , pamh->service_name);
+		return PAM_BUF_ERR;
+	    }
+	    sprintf(filename, PAM_CONFIG_DF, pamh->service_name);
+	    D(("opening %s", filename));
+	    f = fopen(filename, "r");
+	    if (f != NULL) {
+		/* would test magic here? */
+		retval = _pam_parse_conf_file(pamh, f, pamh->service_name);
+		fclose(f);
+		if (retval != PAM_SUCCESS) {
+		    _pam_log_error("_pam_init_handlers: error reading %s"
+				   , filename);
+		    _pam_log_error("_pam_init_handlers: [%s]"
+				   , pam_strerror(retval));
+		} else {
+		    read_something = 1;
+		}
+	    } else {
+		D(("unable to open %s", filename));
+		/*
+		 * XXX - should we log an error? Some people want to always
+		 * use "other"
+		 */
+		retval = PAM_SUCCESS;
+	    }
+	    _pam_drop(filename);
+
+	    if (retval == PAM_SUCCESS) {
+		/* now parse the PAM_DEFAULT_SERVICE_FILE */
+
+		D(("opening %s", PAM_DEFAULT_SERVICE_FILE));
+		f = fopen(PAM_DEFAULT_SERVICE_FILE, "r");
+		if (f != NULL) {
+		    /* would test magic here? */
+		    retval = _pam_parse_conf_file(pamh, f
+						  , PAM_DEFAULT_SERVICE);
+		    fclose(f);
+		    if (retval != PAM_SUCCESS) {
+			_pam_log_error("_pam_init_handlers: error reading %s"
+				       , PAM_DEFAULT_SERVICE_FILE);
+			_pam_log_error("_pam_init_handlers: [%s]"
+				       , pam_strerror(retval));
+		    } else {
+			read_something = 1;
+		    }
+		} else {
+		    D(("unable to open %s", PAM_DEFAULT_SERVICE_FILE));
+		    _pam_log_error("_pam_init_handlers: no default config %s"
+				   , PAM_DEFAULT_SERVICE_FILE);
+		}
+		if (!read_something) {          /* nothing read successfully */
+		    retval = PAM_ABORT;
+		}
+	    }
+	} else {
+	    if ((f = fopen(PAM_CONFIG, "r")) == NULL) {
+		_pam_log_error("_pam_init_handlers: could not open "
+			       PAM_CONFIG );
+		return PAM_ABORT;
+	    }
+
+	    retval = _pam_parse_conf_file(pamh, f, NULL);
+
+	    D(("closing configuration file"));
+	    fclose(f);
+	}
+    }
+
+    if (retval != PAM_SUCCESS) {
 	/* Read error */
-	_pam_log_error("error reading pam.conf");
+	_pam_log_error("error reading PAM configuration file");
 	return PAM_ABORT;
     }
 
@@ -218,7 +322,7 @@ int _pam_init_handlers(pam_handle_t *pamh)
 }
 
 /*
- * This is where we read a line of the pam.conf file. The line may be
+ * This is where we read a line of the PAM config file. The line may be
  * preceeded by lines of comments and also extended with "\\\n"
  */
 
@@ -313,6 +417,7 @@ int _pam_add_handler(pam_handle_t *pamh
     struct handler **handler_p2;
     struct handlers *the_handlers;
     const char *sym, *sym2;
+    char *mod_full_path=NULL;
     servicefn func, func2;
     int success;
 
@@ -322,6 +427,10 @@ int _pam_add_handler(pam_handle_t *pamh
     /* if NULL set to something that can be searched for */
     if (mod_path == NULL) {
 	mod_path = "<*unknown module path*>";
+    } else if (mod_path[0] != '/') {
+	mod_full_path = malloc(sizeof(DEFAULT_MODULE_PATH)+strlen(mod_path));
+	sprintf(mod_full_path, DEFAULT_MODULE_PATH "%s", mod_path);
+	mod_path = mod_full_path;
     }
 
     D(("_pam_add_handler: adding type %d, module `%s'",type,mod_path));
@@ -344,6 +453,7 @@ int _pam_add_handler(pam_handle_t *pamh
 	    if (tmp == NULL) {
 		D(("cannot enlarge module pointer memory"));
 		_pam_log_error("realloc returned NULL in _pam_add_handler");
+		_pam_drop(mod_full_path);
 		return PAM_ABORT;
 	    }
 	    pamh->handlers.module = tmp;
@@ -352,6 +462,7 @@ int _pam_add_handler(pam_handle_t *pamh
 	mod = &(pamh->handlers.module[x]);
 	/* Be pessimistic... */
 	success = PAM_ABORT;
+
 #ifdef PAM_DYNAMIC
 	D(("_pam_add_handler: dlopen(%s) -> %lx", mod_path, &mod->dl_handle));
 	mod->dl_handle = dlopen(mod_path, RTLD_NOW);
@@ -401,16 +512,21 @@ int _pam_add_handler(pam_handle_t *pamh
 	if ((mod->name = _pam_strdup(mod_path)) == NULL) {
 	    D(("_pam_handler: couldn't get memory for mod_path"));
 	    _pam_log_error("no memory for module path", mod_path);
-	    return(PAM_ABORT);
+	    success = PAM_ABORT;
 	}
 
-	/* Now return error if necessary after trying all possible ways... */
-        if (success != PAM_SUCCESS)
-	     return(success);
-    } else {    /* x != pamh->handlers.modules_used */
-	 mod += x;                                   /* the located module */
+    } else {                           /* x != pamh->handlers.modules_used */
+	mod += x;                                    /* the located module */
+	success = PAM_SUCCESS;
     }
 
+    _pam_drop(mod_full_path);
+    mod_path = NULL;                        /* no longer needed or trusted */
+
+    /* Now return error if necessary after trying all possible ways... */
+    if (success != PAM_SUCCESS)
+	return(success);
+
     /*
      * At this point 'mod' points to the stored/loaded module. If its
      * dl_handle is unknown, then we must be able to indicate dispatch
