@@ -1,22 +1,27 @@
 /* pam_mail module */
 
 /*
- * $Id: pam_mail.c,v 1.1 1997/01/04 20:33:02 morgan Exp $
+ * $Id: pam_mail.c,v 1.2 1997/02/15 16:06:14 morgan Exp $
  *
  * Written by Andrew Morgan <morgan@parc.power.net> 1996/3/11
+ * $HOME additions by David Kinchlea <kinch@kinch.ark.com> 1997/1/7
  *
  * $Log: pam_mail.c,v $
+ * Revision 1.2  1997/02/15 16:06:14  morgan
+ * session -> setcred, also added "~"=$HOME
+ *
  * Revision 1.1  1997/01/04 20:33:02  morgan
  * Initial revision
- *
- *
  */
 
 #define DEFAULT_MAIL_DIRECTORY    "/var/spool/mail"
 #define MAIL_FILE_FORMAT          "%s/%s"
-#define MAIL_ENV_FORMAT           "MAIL=%s"
+#define MAIL_ENV_NAME             "MAIL"
+#define MAIL_ENV_FORMAT           MAIL_ENV_NAME "=%s"
 #define YOUR_MAIL_FORMAT          "You have %s mail in %s"
 
+#include <ctype.h>
+#include <pwd.h>
 #include <stdarg.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -26,6 +31,10 @@
 #include <sys/types.h>
 #include <unistd.h>
 
+#ifdef WANT_PWDB
+#include <pwdb/pwdb_public.h>
+#endif
+
 /*
  * here, we make a definition for the externally accessible function
  * in this file (this definition is required for static a module
@@ -33,7 +42,7 @@
  * modules include file to define the function prototypes.
  */
 
-#define PAM_SM_SESSION
+#define PAM_SM_AUTH
 
 #include <security/pam_modules.h>
 #include <security/_pam_macros.h>
@@ -42,9 +51,9 @@
 
 static char *xstrdup(const char *x)
 {
-    char *y;
-
     if (x) {
+	char *y;
+
 	y = (char *) malloc(strlen(x)+1);
 	if (y)
 	    strcpy(y,x);
@@ -55,14 +64,12 @@ static char *xstrdup(const char *x)
 
 static void _log_err(int err, const char *format, ...)
 {
-    char buf[1024];
     va_list args;
 
     va_start(args, format);
-    vsprintf(buf, format, args);
-
-    openlog("pam_mail", LOG_CONS|LOG_PID, LOG_AUTH);
-    syslog(err, buf);
+    openlog("PAM-mail", LOG_CONS|LOG_PID, LOG_AUTH);
+    vsyslog(err, format, args);
+    va_end(args);
     closelog();
 }
 
@@ -74,6 +81,8 @@ static void _log_err(int err, const char *format, ...)
 #define PAM_NEW_MAIL_DIR   010
 #define PAM_MAIL_SILENT    020
 #define PAM_NO_ENV         040
+#define PAM_HOME_MAIL     0100
+#define PAM_EMPTY_TOO     0200
 
 static int _pam_parse(int flags, int argc, const char **argv, char **maildir)
 {
@@ -90,10 +99,10 @@ static int _pam_parse(int flags, int argc, const char **argv, char **maildir)
 
 	if (!strcmp(*argv,"debug"))
 	    ctrl |= PAM_DEBUG_ARG;
-	else if (!strncmp(*argv,"dir=",6)) {
+	else if (!strncmp(*argv,"dir=",4)) {
 	    *maildir = xstrdup(4+*argv);
 	    if (*maildir != NULL) {
-		D(("new mail directory: %s", maildir));
+		D(("new mail directory: %s", *maildir));
 		ctrl |= PAM_NEW_MAIL_DIR;
 	    } else {
 		_log_err(LOG_CRIT,
@@ -105,6 +114,8 @@ static int _pam_parse(int flags, int argc, const char **argv, char **maildir)
 	    ctrl |= PAM_NO_LOGIN;
 	} else if (!strcmp(*argv,"noenv")) {
 	    ctrl |= PAM_NO_ENV;
+	} else if (!strcmp(*argv,"empty")) {
+	    ctrl |= PAM_EMPTY_TOO;
 	} else {
 	    _log_err(LOG_ERR,"pam_parse: unknown option; %s",*argv);
 	}
@@ -153,6 +164,7 @@ static int get_folder(pam_handle_t *pamh, int ctrl
     int retval;
     const char *user, *path;
     char *folder;
+    const struct passwd *pwd=NULL;
 
     retval = pam_get_user(pamh, &user, NULL);
     if (retval != PAM_SUCCESS || user == NULL) {
@@ -162,40 +174,75 @@ static int get_folder(pam_handle_t *pamh, int ctrl
 
     if (ctrl & PAM_NEW_MAIL_DIR) {
 	path = *path_mail;
+	if (*path == '~') {       /* support for $HOME delivery */
+	    pwd = getpwnam(user);
+	    if (pwd == NULL) {
+		_log_err(LOG_ERR, "user [%s] unknown", user);
+		_pam_overwrite(*path_mail);
+		_pam_drop(*path_mail);
+		return PAM_USER_UNKNOWN;
+	    }
+	    /*
+	     * "~/xxx" and "~xxx" are treated as same
+	     */
+	    if (!*++path || (*path == '/' && !*++path)) {
+		_log_err(LOG_ALERT, "badly formed mail path [%s]", *path_mail);
+		_pam_overwrite(*path_mail);
+		_pam_drop(*path_mail);
+		return PAM_ABORT;
+	    }
+	    ctrl |= PAM_HOME_MAIL;
+	}
     } else {
 	path = DEFAULT_MAIL_DIRECTORY;
     }
 
     /* put folder together */
 
-    folder = malloc(sizeof(MAIL_FILE_FORMAT)
-		    +strlen(path)+strlen(user));
-    if (folder == NULL) {
-	_log_err(LOG_CRIT, "out of memory for mail folder");
-	_pam_overwrite(*path_mail);
-	_pam_drop(*path_mail);
-	return PAM_BUF_ERR;
+    if (ctrl & PAM_HOME_MAIL) {
+	folder = malloc(sizeof(MAIL_FILE_FORMAT)
+			+strlen(pwd->pw_dir)+strlen(path));
+    } else {
+	folder = malloc(sizeof(MAIL_FILE_FORMAT)+strlen(path)+strlen(user));
     }
-    sprintf(folder, MAIL_FILE_FORMAT, path, user);
+
+    if (folder != NULL) {
+	if (ctrl & PAM_HOME_MAIL) {
+	    sprintf(folder, MAIL_FILE_FORMAT, pwd->pw_dir, path);
+	} else {
+	    sprintf(folder, MAIL_FILE_FORMAT, path, user);
+	}
+	D(("folder =[%s]", folder));
+    }
+
+    /* tidy up */
 
     _pam_overwrite(*path_mail);
-    _pam_drop(*path_mail);                                   /* tidy up */
+    _pam_drop(*path_mail);
     user = NULL;
 
+    if (folder == NULL) {
+	_log_err(LOG_CRIT, "out of memory for mail folder");
+	return PAM_BUF_ERR;
+    }
+
     *folder_p = folder;
+    folder = NULL;
 
     return PAM_SUCCESS;
 }
 
-static const char *get_mail_status(const char *folder)
+static const char *get_mail_status(int ctrl, const char *folder)
 {
     const char *type;
     struct stat mail_st;
 
     if (stat(folder, &mail_st) == 0 && mail_st.st_size > 0) {
 	type = (mail_st.st_atime < mail_st.st_mtime) ? "new":"old" ;
-    } else {
+    } else if (ctrl & PAM_EMPTY_TOO) {
 	type = "no";
+    } else {
+	type = NULL;
     }
 
     memset(&mail_st, 0, sizeof(mail_st));
@@ -240,21 +287,43 @@ static int report_mail(pam_handle_t *pamh, int ctrl
 
 /* --- authentication management functions (only) --- */
 
+/*
+ * Cannot use mail to authenticate yourself
+ */
+
 PAM_EXTERN
-int pam_sm_open_session(pam_handle_t *pamh,int flags,int argc
-			,const char **argv)
+int pam_sm_authenticate(pam_handle_t *pamh,int flags,int argc
+			 ,const char **argv)
+{
+    return PAM_IGNORE;
+}
+
+/*
+ * MAIL is a "credential"
+ */
+
+PAM_EXTERN
+int pam_sm_setcred(pam_handle_t *pamh, int flags, int argc
+		   , const char **argv)
 {
     int retval, ctrl;
     char *path_mail=NULL, *folder;
     const char *type;
 
     /*
-     * this module sets the MAIL environment variable, and checks if
+     * this module (un)sets the MAIL environment variable, and checks if
      * the user has any new mail.
      */
 
     ctrl = _pam_parse(flags, argc, argv, &path_mail);
 
+    /* Do we have anything to do? */
+
+    flags &= ~PAM_SILENT;
+    if (flags != PAM_CRED_ESTABLISH && flags != PAM_CRED_DELETE) {
+	return PAM_SUCCESS;
+    }
+
     /* which folder? */
 
     retval = get_folder(pamh, ctrl, &path_mail, &folder);
@@ -265,7 +334,7 @@ int pam_sm_open_session(pam_handle_t *pamh,int flags,int argc
 
     /* set the MAIL variable? */
 
-    if (!(ctrl & PAM_NO_ENV)) {
+    if (!(ctrl & PAM_NO_ENV) && (flags & PAM_CRED_ESTABLISH)) {
 	char *tmp;
 
 	tmp = malloc(strlen(folder)+sizeof(MAIL_ENV_FORMAT));
@@ -278,68 +347,38 @@ int pam_sm_open_session(pam_handle_t *pamh,int flags,int argc
 	    if (retval != PAM_SUCCESS) {
 		_pam_overwrite(folder);
 		_pam_drop(folder);
-		_log_err(LOG_CRIT, "unable to set MAIL variable");
+		_log_err(LOG_CRIT, "unable to set " MAIL_ENV_NAME " variable");
 		return retval;
 	    }
 	} else {
-	    _log_err(LOG_CRIT, "no memory for MAIL variable");
+	    _log_err(LOG_CRIT, "no memory for " MAIL_ENV_NAME " variable");
 	    _pam_overwrite(folder);
 	    _pam_drop(folder);
 	    return retval;
 	}
     } else {
-	D(("not setting MAIL variable"));
+	D(("not setting " MAIL_ENV_NAME " variable"));
     }
 
     /*
      * OK. we've got the mail folder... what about its status?
      */
 
-    if (!(ctrl & PAM_NO_LOGIN)) {
-	type = get_mail_status(folder);
-	retval = report_mail(pamh, ctrl, type, folder);
-	type = NULL;
-    }
-
-    _pam_overwrite(folder);                             /* clean up */
-    _pam_drop(folder);
-
-    /* indicate success or failure */
-
-    return retval;
-}
-
-PAM_EXTERN
-int pam_sm_close_session(pam_handle_t *pamh,int flags,int argc
-			 ,const char **argv)
-{
-    int retval, ctrl;
-    char *path_mail=NULL, *folder;
-    const char *type;
-
-    /*
-     * this module sets the MAIL environment variable, and checks if
-     * the user has any new mail.
-     */
-
-    ctrl = _pam_parse(flags, argc, argv, &path_mail);
-
-    /* which folder? */
-
-    retval = get_folder(pamh, ctrl, &path_mail, &folder);
-    if (retval != PAM_SUCCESS) {
-	D(("failed to find folder"));
-	return retval;
+    if ((flags == PAM_CRED_ESTABLISH && !(ctrl & PAM_NO_LOGIN))
+	|| (flags == PAM_CRED_DELETE && ctrl & PAM_LOGOUT_TOO)) {
+	type = get_mail_status(ctrl, folder);
+	if (type != NULL) {
+	    retval = report_mail(pamh, ctrl, type, folder);
+	    type = NULL;
+	}
     }
 
     /*
-     * OK. we've got the mail folder... what about its status?
+     * Delete environment variable?
      */
 
-    if (ctrl & PAM_LOGOUT_TOO) {
-	type = get_mail_status(folder);
-	retval = report_mail(pamh, ctrl, type, folder);
-	type = NULL;
+    if (flags == PAM_CRED_DELETE) {
+	(void) pam_putenv(pamh, MAIL_ENV_NAME);
     }
 
     _pam_overwrite(folder);                             /* clean up */
@@ -356,11 +395,11 @@ int pam_sm_close_session(pam_handle_t *pamh,int flags,int argc
 
 struct pam_module _pam_mail_modstruct = {
      "pam_mail",
+     pam_sm_authenticate,
+     pam_sm_setcred,
      NULL,
      NULL,
      NULL,
-     pam_sm_open_session,
-     pam_sm_close_session,
      NULL,
 };
 
