@@ -1,5 +1,6 @@
 /*----------------------------------------------------------------------
  * Modified for Linux-PAM by Al Longyear <longyear@netcom.com> 96/5/5
+ * Modifications, Cristian Gafton 97/2/8
  *----------------------------------------------------------------------
  * Copyright (c) 1983, 1993, 1994
  *	The Regents of the University of California.  All rights reserved.
@@ -34,7 +35,6 @@
  */
 
 #define USER_RHOSTS_FILE "/.rhosts"     /* prefixed by user's home dir */
-#define THE_SUPERUSER    "root"         /* name should be settable */
 
 #ifdef linux
 #include <endian.h>
@@ -70,6 +70,7 @@
 #include <linux/sockios.h>
 #endif
 
+#define __USE_MISC
 #include <pwd.h>
 #include <sys/file.h>
 #include <sys/signal.h>
@@ -94,6 +95,18 @@ struct _options {
     const char *last_error;
 };
 
+/* logging */
+static void _pam_log(int err, const char *format, ...)
+{
+    va_list args;
+
+    va_start(args, format);
+    openlog("pam_rhosts_auth", LOG_CONS|LOG_PID, LOG_AUTH);
+    vsyslog(err, format, args);
+    va_end(args);
+    closelog();
+}
+
 static void set_option (struct _options *opts, const char *arg)
 {
     if (strcmp (arg, "no_hosts_equiv") == 0) {
@@ -119,6 +132,7 @@ static void set_option (struct _options *opts, const char *arg)
     /*
      * All other options are ignored at the present time.
      */
+    _pam_log(LOG_WARNING, "unrecognized option '%s'", arg);
 }
 
 static void set_parameters (struct _options *opts, int flags,
@@ -304,26 +318,33 @@ iruserok(struct _options *opts, u_long raddr, int superuser,
 	    answer = __ivaliduser(opts, hostf, raddr, luser, ruser);
 	    (void) fclose(hostf);
 	    if (answer == 0)
-		return(0);      /* remote host is equivalent to localhost */
+		return 0;      /* remote host is equivalent to localhost */
 	} /* else {
 	    No hosts.equiv file on system.
 	} */
-
     }
     
     if ( opts->opt_no_rhosts )
-	return (1);
+	return 1;
 
     /*
      * Identify user's local .rhosts file
      */
 
     pwd = getpwnam(luser);
-    if (pwd == NULL)
+    if (pwd == NULL) {
+	/* 
+	 * luser is assumed to be valid because of an earlier check for uid = 0
+	 * we don't log this error twice. However, this shouldn't happen !
+	 * --cristiang 
+	 */
 	return(1);
+    }
 
     /* check for buffer overrun */
     if (strlen(pwd->pw_dir) + sizeof(USER_RHOSTS_FILE) + 2 >= MAXPATHLEN) {
+	if (opts->opt_debug)
+	    _pam_log(LOG_DEBUG,"home directory for `%s' is too long", luser);
 	return 1;                               /* to dangerous to try */
     }
 
@@ -342,7 +363,7 @@ iruserok(struct _options *opts, u_long raddr, int superuser,
     (void)seteuid(uid);
 	
     if (hostf == NULL)
-	return(1);
+	return 1;
 
     /*
      * If not a regular file, or is owned by someone other than
@@ -355,23 +376,21 @@ iruserok(struct _options *opts, u_long raddr, int superuser,
 
     if (fstat(fileno(hostf), &sbuf) < 0)
 	cp = ".rhosts fstat failed";
-    else
-	if (sbuf.st_uid && sbuf.st_uid != pwd->pw_uid)
-	    cp = "bad .rhosts owner";
-	else
-	    if (sbuf.st_mode & (S_IWGRP|S_IWOTH))
-		cp = ".rhosts writable by other than owner";
-
+    else if (sbuf.st_uid && sbuf.st_uid != pwd->pw_uid)
+	cp = "bad .rhosts owner";
+    else if (sbuf.st_mode & (S_IWGRP|S_IWOTH))
+	cp = ".rhosts writable by other than owner";
+    
     /* If there were any problems, quit. */
     if (cp) {
 	opts->last_error = cp;
 	fclose(hostf);
-	return(1);
+	return 1;
     }
 
     answer = __ivaliduser (opts, hostf, raddr, luser, ruser);
     (void) fclose(hostf);
-    return (answer);
+    return answer;
 }
 
 static int
@@ -416,47 +435,73 @@ static int _pam_auth_rhosts (pam_handle_t *pamh,
     const char *luser;
     const char *ruser,*rhost;
     struct _options opts;
-/*
- * Look at the options and set the flags accordingly.
- */
+    int as_root = 0;
+    /*
+     * Look at the options and set the flags accordingly.
+     */
     memset (&opts, 0, sizeof (opts));
     set_parameters (&opts, flags, argc, argv);
-/*
- * Obtain the parameters for the various items
- */
+    /*
+     * Obtain the parameters for the various items
+     */
     for (;;) {                         /* abuse loop to avoid goto */
 
 	/* get the remotehost */
 	retval = pam_get_rhost(pamh, &rhost, NULL);
 	(void) pam_set_item(pamh, PAM_RHOST, rhost);
-	if (retval != PAM_SUCCESS)
+	if (retval != PAM_SUCCESS) {
+	    if (opts.opt_debug) {
+		_pam_log(LOG_DEBUG, "could not get the remote host name");
+	    }
 	    break;
+	}
 
 	/* get the remote user */
 	retval = pam_get_ruser(pamh, &ruser, NULL);
 	(void) pam_set_item(pamh, PAM_RUSER, ruser);
-	if (retval != PAM_SUCCESS)
+	if (retval != PAM_SUCCESS) {
+	    if (opts.opt_debug)
+		_pam_log(LOG_DEBUG, "could not get the remote username");
 	    break;
+	}
 
 	/* get the local user */
 	retval = pam_get_user(pamh, &luser, NULL);
 
-	{
-/* XXX - temporary fix for Red Hat's old libpam-0.50 system */
-	    if (retval == PAM_SUCCESS) {
-		retval = pam_set_item(pamh, PAM_USER, (const void *)luser);
-	    }
+	if (retval != PAM_SUCCESS) {
+	    if (opts.opt_debug)
+		_pam_log(LOG_DEBUG, "could not determine name of local user");
+	    break;
 	}
 
-	if (retval != PAM_SUCCESS)
-	    break;
+	/* check if the luser uid == 0... --cristiang */
+	{
+	    struct passwd *luser_pwd;
+
+	    luser_pwd = getpwnam(luser);
+	    if (luser_pwd == NULL) {
+		if (opts.opt_debug)
+		    _pam_log(LOG_DEBUG, "user '%s' unknown to this system",
+			     luser);
+		retval = PAM_AUTH_ERR;
+		break;
+	    }
+	    if (luser_pwd->pw_uid == 0)
+		as_root = 1;
+	    luser_pwd = NULL;                                   /* forget */
+	}
 /*
  * Validate the account information.
  */
-	if (pam_ruserok (&opts, rhost, !strcmp(THE_SUPERUSER, luser)
-			 , ruser, luser) != 0)
+	if (pam_ruserok (&opts, rhost, as_root, ruser, luser) != 0) {
+	    _pam_log(LOG_WARNING, "denied to %s@%s as %s: %s",
+		     ruser, rhost, luser, (opts.last_error==NULL) ?
+		     "access not allowed":opts.last_error);
 	    retval = PAM_AUTH_ERR;
-
+	} else {
+	    _pam_log(LOG_NOTICE, "allowed to %s@%s as %s",
+                     ruser, rhost, luser);
+	}
 	break;
     }
 
@@ -507,6 +552,10 @@ struct pam_module _pam_rhosts_auth_modstruct = {
 
 /*
  * $Log: pam_rhosts_auth.c,v $
+ * Revision 1.10  1997/02/09 02:09:30  morgan
+ * - implementation of 'debug' argument (Cristian Gafton)
+ * - we check for uid=0 accounts instead of hardcoded 'root' (Cristian Gafton)
+ *
  * Revision 1.9  1996/12/01 03:09:47  morgan
  * *** empty log message ***
  *
